var webusb = {};

(function() {
  'use strict';

  webusb.port = undefined;
  webusb.message = '';
  let textEncoder = new TextEncoder();
  let textDecoder = new TextDecoder();

  let __waitForStr = (str, timeout, clear=true) => new Promise((resolve, reject)=>{
    var check = () => {
      if(webusb.message.indexOf(str) > -1) {
        if (clear)
          webusb.message = '';
        resolve();
      } else if((timeout -= 50) < 0) {
        reject('timeout');
      } else
        setTimeout(check, 50);
    }
    setTimeout(check, 50)
  })

  webusb.connect = async function() {
    webusb.message = '';
    // if port connected, we need to disconnect and connect again
    if (webusb.port) {
      await webusb.port.disconnect().catch(error => {
        console.error(error);
      });
      console.log('disconnect');
    } 
    
    // let user select board
    await webusb.requestPort().then(async (selectedPort) => {
      webusb.port = selectedPort;
      await webusb.port.connect().then(async ()=> {
        // send Ctrl+C twice to interrupt any running program
        let commands = new Uint8Array(1);
        commands[0] = 0x03;
        await webusb.port.send(commands);
        commands[0] = 0x03;
        await webusb.port.send(commands);
        // send Enter to get prompt
        commands[0] = 0x13;
        await this.port.send(commands);

        await __waitForStr('>>>', 1000).catch(() => {return Promise.reject('Response timeout');});
        return Promise.resolve();
      }, error => {
        throw error;
      });
    }).catch(error => {
      console.error('Failed to connect: ');
      console.error(error);
      throw error;
    });
  };

  webusb.connectDefault = async function() {

    // if port connected, we need to disconnect and connect again
    if (webusb.port) {
      await webusb.port.disconnect().catch(error => {
        console.error(error);
      });
    } else { // get default port
      await webusb.getPorts().then(ports => {
        if (ports.length == 0) {
          throw new Error('No device found.');
        } else {
          webusb.port = ports[0];
          webusb.port.connect().then(async ()=> {
            // send Ctrl+C twice to interrupt any running program
            let commands = new Uint8Array(1);
            commands[0] = 0x03;
            await webusb.port.send(commands);
            commands[0] = 0x03;
            await webusb.port.send(commands);
            // send Enter to get prompt
            commands[0] = 0x13;
            await this.port.send(commands);
            
            await __waitForStr('>>>', 1000).catch(() => {return Promise.reject('Response timeout');});
            return Promise.resolve();
          }, error => {
            throw error;
          });
        }}
      ).catch(error => {
        console.error('No port selected');
        console.error(error);
        throw error;
      });
    }
  };

  webusb.disconnect = async function() {
    webusb.message = '';
    if (webusb.port) {
      await webusb.port.disconnect().catch(error => {
        console.error(error);
      });
      webusb.port = undefined;
      console.log('Disconnected');
    } else {
      console.log('No port connected. No need to do anything');
    }
  };

  webusb.sendFile = async function(fileContent, fileName='main.py', append=false, reset=true) {
    if (this.port == undefined) {
      throw new Error("Board not connected");
    }

    // send Ctrl+C twice to interrupt any running program
    let commands = new Uint8Array(1);
    commands[0] = 0x03;
    await this.port.send(commands);
    await this.port.send(commands);

    await __waitForStr('>>>', 2000);

    // send Ctrl+A to enter raw REPL mode
    commands[0] = 0x01;
    await this.port.send(commands);

    await __waitForStr('raw REPL; CTRL-B to exit', 2000);

    // send all code
    let writeMode = '';
    if (append) {
      writeMode = 'a';
    } else {
      writeMode = 'w';
    }
    let commandOpenFile = "f = open('" + fileName + "','" + writeMode + "')\r\n";
    await this.port.send(textEncoder.encode(commandOpenFile)); //textEncoder.encode("f = open('" + fileName + "', " + append?"'a'":"'w'" + ")\r\n"));

    var codeLines = fileContent.split('\n');
    for(var i=0; i<codeLines.length; i++) {
      if (codeLines[i].endsWith('\r') || codeLines[i].endsWith('\n')) {
        codeLines[i] = codeLines[i].slice(0, -1);
      }

      if (codeLines[i].endsWith('\r') || codeLines[i].endsWith('\n')) {
        codeLines[i] = codeLines[i].slice(0, -1);
      }

      // escape special characters like " or \n or \r
      var escapedCode = codeLines[i].replace(/[\\]/g, '\\\\').replace(/["]/g, '\\"');

      var command = 'f.write("' + escapedCode + '\\n")\r\n';

      await this.port.send(textEncoder.encode(command));
    }
    await this.port.send(textEncoder.encode("f.close()\r\n"));

    // send Ctrl+D to save the sent code
    commands[0] = 0x04;
    await this.port.send(commands);

    // send Ctrl+B to exit raw REPL mode
    commands[0] = 0x02;
    await this.port.send(commands);

    await __waitForStr('OK', 2000);

    if (reset) {
      // send Ctrl+D to soft reset
      commands[0] = 0x04;
      await this.port.send(commands);

      await __waitForStr('soft reboot', 2000, false);

      //check if code got error
      await __waitForStr('Traceback', 500).then(() => {
        return Promise.reject('Code bị lỗi không chạy được');
      },
      error => {
        return Promise.resolve();
      });
    } else {
      return Promise.resolve();
    }
    return;
  };

  webusb.getPorts = function() {
    return navigator.usb.getDevices().then(devices => {
      return devices.map(device => new webusb.Port(device));
    });
  };

  webusb.requestPort = function() {
    const filters = [
      { 'vendorId': 0x2341, 'productId': 0x8037 },
    ];
    return navigator.usb.requestDevice({ 'filters': filters }).then(
      device => new webusb.Port(device)
    );
  }

  webusb.processReceivedData = function(data) {
    let data_str = textDecoder.decode(data);
    if (webusb.onReceive) {
      webusb.onReceive(data_str);
    }
    webusb.message += data_str;

    if (webusb.message.length > 1000) {
      // reset when message buffer too large
      webusb.message = '';
    }
  }

  webusb.processReceivedDataError = function(error) {
    if (webusb.onReceiveError) {
      webusb.onReceiveError(error);
    }
  }

  webusb.Port = function(device) {
    this.device_ = device;
    this.interfaceNumber_ = 2;  // original interface number of WebUSB Arduino demo
    this.endpointIn_ = 5;       // original in endpoint ID of WebUSB Arduino demo
    this.endpointOut_ = 4;      // original out endpoint ID of WebUSB Arduino demo
  };

  webusb.Port.prototype.connect = async function() {
    let readLoop = () => {
      this.device_.transferIn(this.endpointIn_, 64).then(result => {
        webusb.processReceivedData(result.data);
        readLoop();
      }, error => {
        webusb.processReceivedDataError(error);
      });
    };

    return this.device_.open()
        .then(() => {
          if (this.device_.configuration === null) {
            return this.device_.selectConfiguration(1);
          }
        })
        .then(() => {
          var configurationInterfaces = this.device_.configuration.interfaces;
          configurationInterfaces.forEach(element => {
            element.alternates.forEach(elementalt => {
              if (elementalt.interfaceClass==0xff) {
                this.interfaceNumber_ = element.interfaceNumber;
                elementalt.endpoints.forEach(elementendpoint => {
                  if (elementendpoint.direction == "out") {
                    this.endpointOut_ = elementendpoint.endpointNumber;
                  }
                  if (elementendpoint.direction=="in") {
                    this.endpointIn_ =elementendpoint.endpointNumber;
                  }
                })
              }
            })
          })
        })
        .then(() => this.device_.claimInterface(this.interfaceNumber_))
        .then(() => this.device_.selectAlternateInterface(this.interfaceNumber_, 0))
        .then(() => this.device_.controlTransferOut({
            'requestType': 'class',
            'recipient': 'interface',
            'request': 0x22,
            'value': 0x01,
            'index': this.interfaceNumber_}))
        .then(() => {
          readLoop();
        });
  };

  webusb.Port.prototype.disconnect = function() {
    return this.device_.controlTransferOut({
            'requestType': 'class',
            'recipient': 'interface',
            'request': 0x22,
            'value': 0x00,
            'index': this.interfaceNumber_})
        .then(() => this.device_.close());
  };

  webusb.Port.prototype.send = function(data) {
    return this.device_.transferOut(this.endpointOut_, data);
  };
})();
