#!/usr/bin/env python3
"""
Build blocks and translations strings.

It's a hack. Needs refactoring!

Example usage:

    $ ./build.py en es de

Will generate new blocks in the "blocks" directory derived from the JSON
definitions in the definitions directory. Translation strings for languages
with the given language codes will be created in the "messages" directory.
These will initially contain the English text defined in the original JSON.
"""
import sys
import json
import copy
import re
from os import listdir, makedirs
from os.path import isfile, join, exists
from string import Template

TRANS_TOKEN = '____{}____'

HEAD = """'use strict';

goog.provide('Blockly.Blocks.colour');
goog.require('Blockly.Blocks');
goog.require('Blockly.constants');

// change color of default categories
Blockly.HSV_SATURATION = 0.65;// 0 (inclusive) to 1 (exclusive), defaulting to 0.45
Blockly.HSV_VALUE = 0.75; // 0 (inclusive) to 1 (exclusive), defaulting to 0.65

// 0-Black, 1-Red, 2-Orange, 3-Yellow, 4-Green, 5-Blue, 6-Indigo, 7-Purple, 8-White
Blockly.FieldColour.COLOURS = ['#FF0000', '#FFA500', '#FFFF00', '#00FF00', '#0000FF', 
                              '#4B0082', '#800080', '#FFFFFF', '#000000'];
Blockly.FieldColour.COLUMNS = 3;

"""

DEFINITION = Template("""Blockly.Blocks['$name'] = {
  init: function() {
    this.jsonInit($definition);
  }
};


""")

EGGSMELL = Template("""<xml id="blockly-toolbox" style="display: none">
$category</xml>
""")

EGGSMELL_CATEGORY = Template("""  <category name="$name" colour="$colour">
$blocks  </category>
""")

EGGSMELL_BLOCK = Template("""    <block type="$name">$shadows</block>\n""")

EGGSMELL_SHADOWS = Template('<value name="$shadow_name">'
                            '<shadow type="$shadow_type">'
                            '<field name="$shadow_item_name">'
                            '$shadow_value'
                            '</field>'
                            '</shadow>'
                            '</value>')

SHADOW_SPECS = {
    'Number': {
        'type': 'math_number',
        'name': 'NUM',
        'value': '0',
    },
    'String': {
        'type': 'text',
        'name': 'TEXT',
        'value': 'Some text',
    },
}

SHADOW_OVERRIDES = {
    'yolobit_led_set_xy': {
        'X': '1',
        'Y': '1'
    },
    'yolobit_led_clear_xy': {
        'X': '1',
        'Y': '1'
    },
    'yolobit_led_set_brightness': {
        'BRIGHTNESS': '10'
    },
    'yolobit_basic_scroll_text': {
        'message': 'Hello, World!',
    },
    'yolobit_basic_sleep': {
        'duration': '1000',
    },
    'yolobit_music_set_tempo': {
        'ticks': '4',
        'bpm': '120',
    },
    'yolobit_music_pitch': {
        'pitch': '440',
        'duration': '1000',
    },
    'yolobit_math_map': {
        'VALUE': '0',
        'FROM_LOW': '0',
        'FROM_HIGH': '1023',
        'TO_LOW': '0',
        'TO_HIGH': '100',
    },
    'yolobit_wifi_map': {
        'WIFI': 'wifi',
        'PASSWORD': 'password',
        'NAME': 'my_yolobit_123',
        'MESSAGE': 'hello',
        'NAME': 'other_yolobit_456',
        'DATA': '123',
        'TOPIC': 'topic1'
    }
}


LANGUAGE_HEAD = Template("""'use strict';

goog.provide('Blockly.Msg.$language');

goog.require('Blockly.Msg');


""")

CATEGORIES = [
    {'name': 'basic', 'colour': '#1e90ff'},
    {'name': 'input', 'colour': '#d400d4'},
    {'name': 'music', 'colour': '#e63022'},
    {'name': 'led', 'colour': '#5d2c91'},
    {'name': 'pins', 'colour': '#b22222'},
    {'name': 'math', 'colour': '#4257bf'},
    {'name': 'wifi', 'colour': '#e3008c'}
]


def get_categories(path='definitions'):
    """
    Return data structures representing the various block definitions.
    """
    files = sorted([f for f in listdir(path) if
             isfile(join(path, f)) and f.endswith('.js')])
    result = []
    for c in CATEGORIES:
        with open(join(path, c['name'] + '.js')) as def_file:
            result.append({
              'name': c['name'], #f.replace('.js', '').capitalize(),
              'colour': c['colour'],
              'definitions': json.loads(def_file.read())
            })
    """
    for f in files:
        with open(join(path, f)) as def_file:
            result.append({
              'name': f.replace('.js', '').capitalize(),
              'definitions': json.loads(def_file.read())
            })
    """
    return result


def get_code(definitions, colour):
    """
    Given a list of definitions will return the associated code.
    """
    code = []
    for definition in definitions:
        new_def = copy.deepcopy(definition)
        name = new_def['type']
        del new_def['type']
        new_def['colour'] = colour
        code.append(DEFINITION.substitute(name=name,
                                          definition=json.dumps(new_def)))
    code = ''.join(code)
    matcher = re.compile(r'\"____(?P<name>[\w]+)____\"')
    matched = matcher.findall(code)
    for name in matched:
        code = code.replace('"' + TRANS_TOKEN.format(name) + '"',
                            'Blockly.Msg.{}'.format(name))
    return code


def get_eggsmell(categories):
    """
    Gets XML (pronounced eggsmell) for toolbox. Seriously, this is far quicker
    than faffing about with XML.
    """
    rendered = []
    for category in categories:
        rendered_defs = []
        for definition in category['definitions']:
            name = definition['type']
            shadows = ''
            for arg in definition.get('args0', []):
                if arg['type'] == 'input_value' and \
                ('check' in arg and arg['check'] in SHADOW_SPECS):
                    spec = SHADOW_SPECS[arg['check']]
                    s_name = arg['name']
                    s_type = spec['type']
                    s_item_name = spec['name']
                    s_val = spec['value']
                    if name in SHADOW_OVERRIDES:
                        if s_name in SHADOW_OVERRIDES[name]:
                            s_val = SHADOW_OVERRIDES[name][s_name]
                    shadows += EGGSMELL_SHADOWS.substitute(shadow_name=s_name,
                                                           shadow_type=s_type,
                                                           shadow_item_name=s_item_name,
                                                           shadow_value=s_val)
            rendered_defs.append(EGGSMELL_BLOCK.substitute(name=name,
                                                           shadows=shadows))
        blocks = ''.join(rendered_defs)
        rendered.append(EGGSMELL_CATEGORY.substitute(name=category['name'],
                                                     colour=category['colour'],
                                                     blocks=blocks))
    return EGGSMELL.substitute(category=''.join(rendered))


def generate_translation(definition):
    """
    Given a JSON representation of a definition will return a list of
    JavaScript statements defining the attributes to be translated.
    """
    result = []
    fields = ['tooltip', 'helpUrl']
    template = "Blockly.Msg.{} = {};\n"
    name = definition['type']
    for k in definition.keys():
        trans_name = '{}_{}'.format(name, k).upper()
        if k in fields:
            result.append(template.format(trans_name, repr(definition[k])))
            definition[k] = TRANS_TOKEN.format(trans_name)
        elif k.startswith('message'):
            result.append(template.format(trans_name, repr(definition[k])))
            definition[k] = TRANS_TOKEN.format(trans_name)
    return result


def main(languages):
    categories = get_categories()
    colour = 0
    step = 360 // len(categories)
    #for i, category in enumerate(categories):
        #category['colour'] = colour
        #colour = step * (i + 1)
    for lang in languages:
        lang_content = []
        lang_content.append(LANGUAGE_HEAD.substitute(language=lang))
        for category in categories:
            for definition in category['definitions']:
                lang_content.extend(generate_translation(definition))
        translation_strings = ''.join(lang_content)
        directory = join('messages', lang)
        if not exists(directory):
            mkdirs(directory)
        with open(join(directory, 'messages.js'), 'w') as msgs:
            msgs.write(translation_strings)
            print('Done\n')

    with open('blocks/yolobit.js', 'w') as output:
        output.write(HEAD)
        for category in categories:
            output.write(get_code(category['definitions'], category['colour']))
    with open('toolbox.xml', 'w') as toolbox:
        toolbox.write(get_eggsmell(categories))


if __name__ == '__main__':
    main(sys.argv[1:])
